==> sems-0.8.12.tar.gz: sems-0.8.12/AmSession.cpp <==
/*
 * $Id: AmSession.cpp,v 1.31.2.6 2005/08/25 07:54:05 rco Exp $
 *
 * Copyright (C) 2002-2003 Fhg Fokus
 *
 * This file is part of sems, a free SIP media server.
 *
 * sems is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * sems is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "AmServer.h"
#include "AmSession.h"
#include "AmRequest.h"
#include "AmSdp.h"
#include "AmMail.h"
#include "AmConfig.h"
#include "AmUtils.h"
#include "AmPlugIn.h"
#include "AmApi.h"

#include "log.h"

#include <algorithm>

#include <unistd.h>
#include <assert.h>

AmSessionEvent::AmSessionEvent(EventType event_type, const AmRequestUAS& request)
    : AmEvent(int(event_type)), request(request)
{
}

// AmSession methods

AmSession::AmSession(AmRequest* _req, AmDialogState* _d_st)
    : req(_req), dialog_st(_d_st), sess_stopped(false),
      rtp_str(this)
{
}

AmSession::~AmSession()
{
    delete dialog_st;
}

void AmSession::onIdle()
{
    dialog_st->processEvents();
}

const string& AmSession::getCallID() 
{ 
    return req->cmd.callid; 
}

const string& AmSession::getFromTag() 
{ 
    return req->cmd.from_tag; 
}

const string& AmSession::getToTag()
{
    return req->cmd.to_tag;
}

AmDialogState* AmSession::getDialogState()
{
  return dialog_st;
}

const SdpPayload& AmSession::getPayload()
{
    return payload;
}

int AmSession::getRPort()
{
    return rtp_str.getRPort();
}

void AmSession::run()
{
    if(dialog_st){

	try {
	    try {

		string r_host;
		int    r_port = 0;
		SdpMedia::Direction dir = SdpMedia::DirBoth;
	    
		if(req->sdp.parse())
		    throw AmSession::Exception(400,"session description parsing failed");

		if(req->sdp.media.empty())
		    throw AmSession::Exception(400,"no media line found in SDP message");
	    
		if(req->sdp.getCompatiblePayload(MT_AUDIO,payload,r_host,r_port,dir))
		    throw AmSession::Exception(606,"could not find compatible payload.");

		string str_msg_flags = req->cmd.getHeader("P-MsgFlags");
		unsigned int msg_flags;
		if(reverse_hex2int(str_msg_flags,msg_flags)){
		  ERROR("while parsing 'P-MsgFlags' header\n");
		  msg_flags = 0;
		}

		bool passive_mode=false;
		DBG("dir=%i; msg_flags=%u\n",dir,msg_flags);
		if((dir == SdpMedia::DirActive)
		   || (msg_flags & FL_FORCE_ACTIVE)) {
		    DBG("The other UA is NATed: switched to passive mode.\n");
		    passive_mode = true;
		}

		if(!req->getReplied()){
		    dialog_st->onBeforeCallAccept(req.get());
		    rtp_str.setLocalIP(req->cmd.dstip);
		    req->accept(rtp_str.getLocalPort(),payload,dir);
		}

		rtp_str.setPassiveMode(passive_mode);
		rtp_str.setRAddr(r_host,r_port);
		DBG("Sending Rtp data to %s/%i\n",r_host.c_str(),r_port);

		// enable RTP stream
		rtp_str.start(payload);

		dialog_st->onSessionStart(req.get());
		if(!req->finished && !getStopping())
		    req->bye();
	    }
	    catch(const string& str){
		ERROR("%s\n",str.c_str());
		throw AmSession::Exception(500,"unexpected exception.");
	    }
	    catch(const AmSession::Exception& e){ throw e; }
	    catch(...){
		throw AmSession::Exception(500,"unexpected exception.");
	    }
	}
	catch(const AmSession::Exception& e){
	    ERROR("%i %s\n",e.code,e.reason.c_str());
	    dialog_st->onError(e.code,e.reason);
	    req->error(e.code,e.reason);
	}
    }

    destroy();
}

void AmSession::on_stop()
{
    sess_stopped.set(true);
    rtp_str.pause();

    if(dialog_st)
        dialog_st->onBye(req.get());

    // just to wake up threads 
    // waiting for events.
    dialog_st->postEvent(0);
}


bool AmSession::modify(AmRequest* new_req)
{
    // we don't accept session modification yet...
    DBG("sorry, we don't accept re-invites\n");
    new_req->error(488,"Not Acceptable Here");
    return true;
}

void AmSession::destroy()
{
    AmSessionContainer::instance()->destroySession(this);
}

string AmSession::getNewId()
{
    return int2hex(getpid()) + int2hex(rand());
}

// AmSessionContainer methods

AmSessionContainer* AmSessionContainer::_SessionContainer=0;

AmSessionContainer::AmSessionContainer()
    : _run_cond(false)
{
}

AmSessionContainer* AmSessionContainer::instance()
{
    if(!_SessionContainer)
	_SessionContainer = new AmSessionContainer();

    return _SessionContainer;
}

void AmSessionContainer::on_stop() 
{ 
}

void AmSessionContainer::run()
{
    int seed=0;
    FILE* fp_rand = fopen("/dev/random","r");
    if(fp_rand){
	fread(&seed,sizeof(int),1,fp_rand);
	fclose(fp_rand);
    }
    seed += getpid();
    seed += time(0);
    srand(seed);

    while(1){

	_run_cond.wait_for();

	// Let some time for the Sessions 
	// to stop by themselves
	sleep(5);

	ds_mut.lock();
	DBG("Session cleaner starting its work\n");
	
	try {
	    SessionQueue n_sessions;

	    while(!d_sessions.empty()){

	        AmSession* cur_session = d_sessions.front();
		d_sessions.pop();

		ds_mut.unlock();

		if(cur_session->getStopped()){
		    delete cur_session;
		    DBG("session %p has been destroyed'\n",cur_session);
		}
		else {
		    DBG("session %p still running\n",cur_session);
		    n_sessions.push(cur_session);
		}

		ds_mut.lock();
	    }

	    swap(d_sessions,n_sessions);

	}catch(std::exception& e){
		ERROR("exception caught in session cleaner: %s\n", e.what());
		throw; /* throw again as this is fatal (because unlocking the mutex fails!! */
	}catch(...){
		ERROR("unknown exception caught in session cleaner!\n");
		throw; /* throw again as this is fatal (because unlocking the mutex fails!! */
	}

	bool more = !d_sessions.empty();
	ds_mut.unlock();

	DBG("Session cleaner finished\n");
	if(!more)
	    _run_cond.set(false);
    }
}

void AmSessionContainer::stopAndQueue(AmSession* s)
{
    ds_mut.lock();

    DBG("session cleaner trying to stop %p\n",s);
    s->stop();
    d_sessions.push(s);
    _run_cond.set(true);
    
    ds_mut.unlock();
}

void AmSessionContainer::destroySession(AmSession* s)
{
    string hash_str = string(s->getCallID()) + string(s->getFromTag());
    string to_tag = s->getToTag();
    sadSession(hash_str,to_tag);
}

bool AmSessionContainer::sadSession(const string& hash_str, string& sess_key)
{
    bool ret;
    as_mut.lock();
    try {
	DBG("searching for session to destroy (hash=%s,sess_key=%s)\n",hash_str.c_str(),sess_key.c_str());
	pair<SessionMapIter,SessionTableIter> sess_pair = findSession(hash_str,sess_key);
	ret = (sess_pair.first != a_sessions.end())
	    && (sess_pair.second != sess_pair.first->second.end());
	destroySession(sess_pair,true);
    }catch(...){
	ERROR("unexpected exception\n");
		}
    as_mut.unlock();
    return ret;
}

void AmSessionContainer::destroySession(pair<SessionMapIter,SessionTableIter>& sess_pair, bool sessionsAreLocked)
{
    if (!sessionsAreLocked) as_mut.lock();

    try {
	SessionMapIter sess_it = sess_pair.first;
	SessionTableIter it = sess_pair.second;
	
	if(sess_it != a_sessions.end()){
	    
	    DBG("sess_it != a_sessions.end()\n");
	    if(it != sess_it->second.end()){
		
		DBG("it != sess_it->second.end()\n");
		
		AmSession* am_sess = it->second;
		sess_it->second.erase(it);
		if(!sess_it->second.size())
		    a_sessions.erase(sess_it);
		
		DBG("session found: stopping session\n");
		stopAndQueue(am_sess);
		DBG("session stopped and destroyed (#sessions=%i)\n",(int)a_sessions.size());
	    }
	    else {
		DBG("it == sess_it->second.end()\n");
	    }
	}
	else {
	    DBG("sess_it == sessions.end()\n");
	}
    }catch(...){
	ERROR("unexpected exception\n");
		}

    if (!sessionsAreLocked) as_mut.unlock();
}


AmSessionContainer::SessionSearchRes 
AmSessionContainer::findSession(const string& hash_str, string& sess_key)
{
    SessionTableIter it;
    SessionMapIter sess_it = a_sessions.find(hash_str);

    if(sess_it != a_sessions.end()){

	if(sess_key.length()){
	    
	    for(it=sess_it->second.begin();it!=sess_it->second.end();++it){
		
		if(it->first == sess_key){
		    DBG("session found with key\n");
		    break;
		}
	    }
	}
	else if(!sess_it->second.empty()){
	    DBG("sess_key is null: using the last session key\n");
	    it = sess_it->second.end()-1;
	    sess_key = it->first;
	}
	else
	    it = sess_it->second.end();
    }

    return make_pair(sess_it,it);
}

AmSession* AmSessionContainer::getSession(const string& hash_str, string& sess_key)
{
    pair<SessionMapIter,SessionTableIter> sess_pair = findSession(hash_str,sess_key);
	
    if( (sess_pair.first != a_sessions.end()) 
	&& (sess_pair.second != sess_pair.first->second.end()) ){
	    
	return sess_pair.second->second;
    }

    return NULL;
}

void AmSessionContainer::startSession(const string& hash_str, string& sess_key,
				      AmRequestUAS* req)
{
    as_mut.lock();
    try {

	AmSession* session = AmSessionContainer::getSession(hash_str,sess_key);
	if( session ){
	    
	    DBG("AmSessionContainer: session already exists.\n");
	    // Session already exists...
	    if(!req->cmd.to_tag.empty()){
		// ...it's a RE-INVITE
		session->modify(req);
	    }
	    else {
		//DBG("current local tag: %s\n",session->req->cmd.to_tag.c_str());
		req->cmd.to_tag = session->req->cmd.to_tag;
		// it's a forked-and-merged INVITE
		// reply 482 Loop detected
		throw AmSession::Exception(482, "Loop detected");
	    }
	}
	else {

	    // Call-ID and From-Tag are unknown: it's a new session

	    sess_key = AmSession::getNewId();
	    AmSession* session = createSession(req->cmd.cmd,req);

 	    if(session && !addSession(hash_str,sess_key,session)){
 		DBG( "Starting session... (hash=%s,sess_key=%s)\n",
 		     hash_str.c_str(),req->cmd.to_tag.c_str());
 		session->start();
 	    }
	}
    } 
    catch(const AmSession::Exception& e){
 	ERROR("%i %s\n",e.code,e.reason.c_str());
 	req->reply(e.code,e.reason);
    }
    catch(const string& err){
	ERROR("startSession: %s\n",err.c_str());
	req->reply(500,err);
    }
    catch(...){
	ERROR("unexpected exception\n");
		}
    as_mut.unlock();
}

bool AmSessionContainer::postEvent(const string& hash_str,string& sess_key,
				   AmSessionEvent* event)
{
    bool ret = false;
    as_mut.lock();
    try {
	AmSession* session = AmSessionContainer::getSession(hash_str,sess_key);
	if( session ){
	    AmDialogState* state = session->getDialogState();
	    assert(state);
	    state->postEvent(event);
	    ret = true;
	}
    } catch(...){
	ERROR("unexpected exception\n");
		}
    as_mut.unlock();
    return ret;
}

AmSession* AmSessionContainer::createSession(const string& app_name, AmRequest* req)
{
    AmStateFactory* state_factory = AmPlugIn::instance()->getFactory4App(app_name);
    if(!state_factory) {
	ERROR("application '%s' not found !", app_name.c_str());
	throw string("application '" + app_name + "' not found !");
		}
	    
    AmDialogState* st = state_factory->onInvite(req->cmd);
    if(!st) {
	//  State creation failed:
	//   application denied session creation
	//   or there was an error.
	//
	//  let's hope the createState function has replied...
	//  ... and do nothing !

	return 0;
    }

    // create a new session
    st->state_factory = state_factory;
    AmSession* session = new AmSession(req->duplicate(),st);
    st->session = session;
    return session;
}

bool AmSessionContainer::addSession(const string& hash_str,const string& sess_key,
				    AmSession* session)
{
    SessionMapIter sess_it = a_sessions.find(hash_str);
    bool ret = sess_it != a_sessions.end();
    a_sessions[hash_str].push_back(make_pair(sess_key,session));
    return ret;
}

int AmSessionContainer::getSize()
{
    int res=0;
    as_mut.lock();
    res = a_sessions.size();
    as_mut.unlock();
    return res;
}

==> sems-0.9.0.tar.gz: sems-0.9.0/AmSession.cpp <==
/*
 * $Id: AmSession.cpp,v 1.43.2.1 2005/08/25 09:34:20 rco Exp $
 *
 * Copyright (C) 2002-2003 Fhg Fokus
 *
 * This file is part of sems, a free SIP media server.
 *
 * sems is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * sems is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "AmServer.h"
#include "AmSession.h"
#include "AmRequest.h"
#include "AmSdp.h"
#include "AmMail.h"
#include "AmConfig.h"
#include "AmUtils.h"
#include "AmPlugIn.h"
#include "AmApi.h"

#include "log.h"

#include <algorithm>

#include <unistd.h>
#include <assert.h>
AmSessionEvent::AmSessionEvent(EventType event_type, const AmRequestUAS& request)
    : AmEvent(int(event_type)), request(request)
{
}

AmNotifySessionEvent::AmNotifySessionEvent(AmRequestUAS& request)
    : AmSessionEvent(AmSessionEvent::Notify, request),
      body(request.getBody()),
      eventPackage(request.cmd.getHeader("Event")),
      subscriptionState(request.cmd.getHeader("Subscription-State")),
      contentType(request.cmd.getHeader("Content-Type")),
      contentLength(request.cmd.getHeader("Content-Length")),
      cseq(request.cmd.cseq)
{
}

// AmSession methods

AmSession::AmSession(AmRequest* _req, AmDialogState* _d_st)
    : req(_req), dialog_st(_d_st), sess_stopped(false),
      rtp_str(this),first_negotiation(true)
{
}

AmSession::~AmSession()
{
    delete dialog_st;
}

void AmSession::onIdle()
{
    dialog_st->processEvents();
}

const string& AmSession::getCallID() 
{ 
    return req->cmd.callid; 
}

const string& AmSession::getFromTag() 
{ 
    return req->cmd.from_tag; 
}

const string& AmSession::getToTag()
{
    return req->cmd.to_tag;
}

AmDialogState* AmSession::getDialogState()
{
  return dialog_st;
}

const SdpPayload& AmSession::getPayload()
{
    return payload;
}

int AmSession::getRPort()
{
    return rtp_str.getRPort();
}

void AmSession::negotiate(AmRequest* request)
{
    string r_host;
    int    r_port = 0;
    SdpMedia::Direction dir = SdpMedia::DirBoth;
    SdpPayload tmp_pl;
    
    if(request->sdp.parse())
	throw AmSession::Exception(400,"session description parsing failed");
    
    if(request->sdp.media.empty())
	throw AmSession::Exception(400,"no media line found in SDP message");
    
    if(request->sdp.getCompatiblePayload(MT_AUDIO,tmp_pl,r_host,r_port,dir))
	throw AmSession::Exception(606,"could not find compatible payload");
    
    if(first_negotiation){
	payload = tmp_pl;
    }
    else if(payload.int_pt != tmp_pl.int_pt)
	throw AmSession::Exception(400,"do not accept payload changes");

    string str_msg_flags = request->cmd.getHeader("P-MsgFlags");
    unsigned int msg_flags;
    if(reverse_hex2int(str_msg_flags,msg_flags)){
	ERROR("while parsing 'P-MsgFlags' header\n");
	msg_flags = 0;
    }
    DBG("dir=%i; msg_flags=%u\n",dir,msg_flags);
    
    bool passive_mode = false;
    if( (dir == SdpMedia::DirActive)
	|| (msg_flags & FL_FORCE_ACTIVE) ) {
	DBG("The other UA is NATed: switched to passive mode.\n");
	passive_mode = true;
    }
    
    if(!request->getReplied()){
	
	//TODO: if dialog started
	if(first_negotiation)
	    dialog_st->onBeforeCallAccept(request);
	
	if (sess_stopped.get()) /* session still alive ? */
	    return;
	
	rtp_str.setLocalIP(request->cmd.dstip);
	request->accept(rtp_str.getLocalPort(),payload,dir);
    }
    
    rtp_str.setPassiveMode(passive_mode);
    rtp_str.setRAddr(r_host, r_port);
    DBG("Sending Rtp data to %s/%i\n",r_host.c_str(),r_port);
    
    first_negotiation = false;
}

void AmSession::run()
{
    if(dialog_st){

	try {
	    try {
		// handle INVITE negotiation
		negotiate(req.get());

		if(sess_stopped.get()){
		    destroy();
		    return;
		}

		// enable RTP stream
		rtp_str.start(payload);
		dialog_st->onSessionStart(req.get());
	    }
	    catch(const AmSession::Exception& e){ throw e; }
	    catch(const string& str){
		ERROR("%s\n",str.c_str());
		throw AmSession::Exception(500,"unexpected exception.");
	    }
	    catch(...){
		throw AmSession::Exception(500,"unexpected exception.");
	    }
	}
	catch(const AmSession::Exception& e){
	    ERROR("%i %s\n",e.code,e.reason.c_str());
	    dialog_st->onError(e.code,e.reason);
	    req->error(e.code,e.reason);
	}
    }

    destroy();
}

void AmSession::on_stop()
{
    sess_stopped.set(true);
    rtp_str.pause();

    if(dialog_st)
        dialog_st->onBye(req.get());

    // just to wake up threads 
    // waiting for events.
    dialog_st->postEvent(0);
}

void AmSession::destroy()
{
    AmSessionContainer::instance()->destroySession(this);
}

string AmSession::getNewId()
{
    return int2hex(getpid()) + int2hex(rand());
}

// AmSessionContainer methods

AmSessionContainer* AmSessionContainer::_SessionContainer=0;

AmSessionContainer::AmSessionContainer()
    : _run_cond(false)
{
}

AmSessionContainer* AmSessionContainer::instance()
{
    if(!_SessionContainer)
	_SessionContainer = new AmSessionContainer();

    return _SessionContainer;
}

void AmSessionContainer::on_stop() 
{ 
}

void AmSessionContainer::run()
{
    int seed=0;
    FILE* fp_rand = fopen("/dev/random","r");
    if(fp_rand){
	fread(&seed,sizeof(int),1,fp_rand);
	fclose(fp_rand);
    }
    seed += getpid();
    seed += time(0);
    srand(seed);

    while(1){

	_run_cond.wait_for();

	// Let some time for the Sessions 
	// to stop by themselves
	sleep(5);

	ds_mut.lock();
	DBG("Session cleaner starting its work\n");
	
	try {
	    SessionQueue n_sessions;

	    while(!d_sessions.empty()){

	        AmSession* cur_session = d_sessions.front();
		d_sessions.pop();

		ds_mut.unlock();

		if(cur_session->getStopped()){
		    delete cur_session;
		    DBG("session %ld has been destroyed'\n",(unsigned long int)cur_session->_pid);
		}
		else {
		    DBG("session %ld still running\n",(unsigned long int)cur_session->_pid);
		    n_sessions.push(cur_session);
		}

		ds_mut.lock();
	    }

	    swap(d_sessions,n_sessions);

	}catch(std::exception& e){
		ERROR("exception caught in session cleaner: %s\n", e.what());
		throw; /* throw again as this is fatal (because unlocking the mutex fails!! */
	}catch(...){
		ERROR("unknown exception caught in session cleaner!\n");
		throw; /* throw again as this is fatal (because unlocking the mutex fails!! */
	}

	bool more = !d_sessions.empty();
	ds_mut.unlock();

	DBG("Session cleaner finished\n");
	if(!more)
	    _run_cond.set(false);
    }
}

void AmSessionContainer::stopAndQueue(AmSession* s)
{
    ds_mut.lock();

    DBG("session cleaner trying to stop %ld\n",(unsigned long int)s->_pid);
    s->stop();
    d_sessions.push(s);
    _run_cond.set(true);
    
    ds_mut.unlock();
}

void AmSessionContainer::destroySession(AmSession* s)
{
    string hash_str = string(s->getCallID()) + string(s->getFromTag());
    string to_tag = s->getToTag();
    sadSession(hash_str,to_tag);
}

bool AmSessionContainer::sadSession(const string& hash_str, string& sess_key)
{
    bool ret;
    as_mut.lock();
    try {
	DBG("searching for session to destroy (hash=%s,sess_key=%s)\n",hash_str.c_str(),sess_key.c_str());
	pair<SessionMapIter,SessionTableIter> sess_pair = findSession(hash_str,sess_key);
	ret = (sess_pair.first != a_sessions.end())
	    && (sess_pair.second != sess_pair.first->second.end());
	destroySession(sess_pair,true);
    }catch(...){
	ERROR("unexpected exception\n");
		}
    as_mut.unlock();
    return ret;
}

void AmSessionContainer::destroySession(pair<SessionMapIter,SessionTableIter>& sess_pair, bool sessionsAreLocked)
{
    if (!sessionsAreLocked) as_mut.lock();

    try {
	SessionMapIter sess_it = sess_pair.first;
	SessionTableIter it = sess_pair.second;
	
	if(sess_it != a_sessions.end()){
	    
	    DBG("sess_it != a_sessions.end()\n");
	    if(it != sess_it->second.end()){
		
		DBG("it != sess_it->second.end()\n");
		
		AmSession* am_sess = it->second;
		sess_it->second.erase(it);
		if(!sess_it->second.size())
		    a_sessions.erase(sess_it);
		
		DBG("session found: stopping session\n");
		stopAndQueue(am_sess);
		DBG("session stopped and destroyed (#sessions=%li)\n",a_sessions.size());
	    }
	    else {
		DBG("it == sess_it->second.end()\n");
	    }
	}
	else {
	    DBG("sess_it == sessions.end()\n");
	}
    }catch(...){
	ERROR("unexpected exception\n");
		}

    if (!sessionsAreLocked) as_mut.unlock();
}


AmSessionContainer::SessionSearchRes 
AmSessionContainer::findSession(const string& hash_str, string& sess_key)
{
    SessionTableIter it;
    SessionMapIter sess_it = a_sessions.find(hash_str);

    if(sess_it != a_sessions.end()){

	if(sess_key.length()){
	    
	    for(it=sess_it->second.begin();it!=sess_it->second.end();++it){
		
		if(it->first == sess_key){
		    DBG("session found with key\n");
		    break;
		}
	    }
	}
	else if(!sess_it->second.empty()){
	    DBG("sess_key is null: using the last session key\n");
	    it = sess_it->second.end()-1;
	    sess_key = it->first;
	}
	else
	    it = sess_it->second.end();
    }

    return make_pair(sess_it,it);
}

AmSession* AmSessionContainer::getSession(const string& hash_str, string& sess_key)
{
    pair<SessionMapIter,SessionTableIter> sess_pair = findSession(hash_str,sess_key);
	
    if( (sess_pair.first != a_sessions.end()) 
	&& (sess_pair.second != sess_pair.first->second.end()) ){
	    
	return sess_pair.second->second;
    }

    return NULL;
}

void AmSessionContainer::startSession(const string& hash_str, string& sess_key,
				      AmRequestUAS* req)
{
    as_mut.lock();
    try {

	AmSession* session = AmSessionContainer::getSession(hash_str,sess_key);
	if( session ){
	    
	    DBG("AmSessionContainer: session already exists.\n");
	    // Session already exists...
	    if(!req->cmd.to_tag.empty()){
		// ...it's a RE-INVITE
		AmDialogState* state = session->getDialogState();
		assert(state);
		state->postEvent(new AmSessionEvent(AmSessionEvent::ReInvite,
						    *req));
	    }
	    else {
		req->cmd.to_tag = session->req->cmd.to_tag;
		// it's a forked-and-merged INVITE
		// reply 482 Loop detected
		throw AmSession::Exception(482, "Loop detected");
	    }
	}
	else {

	    // Call-ID and From-Tag are unknown: it's a new session

	    sess_key = AmSession::getNewId();
	    AmSession* session = createSession(req->cmd.cmd,req);

 	    if(session && !addSession(hash_str,sess_key,session)){
 		DBG( "Starting session... (hash=%s,sess_key=%s)\n",
 		     hash_str.c_str(),req->cmd.to_tag.c_str());
 		session->start();
 	    }
	}
    } 
    catch(const AmSession::Exception& e){
 	ERROR("%i %s\n",e.code,e.reason.c_str());
 	req->reply(e.code,e.reason);
    }
    catch(const string& err){
	ERROR("startSession: %s\n",err.c_str());
	req->reply(500,err);
    }
    catch(...){
	ERROR("unexpected exception\n");
		}
    as_mut.unlock();
}

bool AmSessionContainer::postEvent(const string& hash_str,string& sess_key,
				   AmSessionEvent* event)
{
    bool ret = false;
    as_mut.lock();
    try {
	AmSession* session = AmSessionContainer::getSession(hash_str,sess_key);
	if( session ){
	    AmDialogState* state = session->getDialogState();
	    assert(state);
	    state->postEvent(event);
	    ret = true;
	}
    } catch(...){
	ERROR("unexpected exception\n");
		}
    as_mut.unlock();
    return ret;
}

AmSession* AmSessionContainer::createSession(const string& app_name, AmRequest* req)
{
    AmStateFactory* state_factory = AmPlugIn::instance()->getFactory4App(app_name);
    if(!state_factory) {
	ERROR("application '%s' not found !", app_name.c_str());
	throw string("application '" + app_name + "' not found !");
		}
	    
    AmDialogState* st = state_factory->onInvite(req->cmd);
    if(!st) {
	//  State creation failed:
	//   application denied session creation
	//   or there was an error.
	//
	//  let's hope the createState function has replied...
	//  ... and do nothing !

	return 0;
    }

    // create a new session
    st->state_factory = state_factory;
    AmSession* session = new AmSession(req->duplicate(),st);
    st->session = session;
    return session;
}

bool AmSessionContainer::addSession(const string& hash_str,const string& sess_key,
				    AmSession* session)
{
    SessionMapIter sess_it = a_sessions.find(hash_str);
    bool ret = sess_it != a_sessions.end();
    a_sessions[hash_str].push_back(make_pair(sess_key,session));
    return ret;
}

int AmSessionContainer::getSize()
{
    int res=0;
    as_mut.lock();
    res = a_sessions.size();
    as_mut.unlock();

    return res;
}
